#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ....operands import SwapAxes
from ..utils import validate_axis
from ..core import TensorOperandMixin


def _swap(it, axis1, axis2):
    new_it = list(it)
    new_it[axis1], new_it[axis2] = it[axis2], it[axis1]

    return tuple(new_it)


class TensorSwapAxes(SwapAxes, TensorOperandMixin):
    def __init__(self, axis1=None, axis2=None, dtype=None, sparse=False, **kw):
        super(TensorSwapAxes, self).__init__(_axis1=axis1, _axis2=axis2, _dtype=dtype,
                                             _sparse=sparse, **kw)

    def __call__(self, a):
        shape = _swap(a.shape, self.axis1, self.axis2)
        return self.new_tensor([a], shape)

    def _set_inputs(self, inputs):
        super(TensorSwapAxes, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        axis1, axis2 = op.axis1, op.axis2
        in_tensor = op.inputs[0]

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_shape = _swap(c.shape, axis1, axis2)
            chunk_idx = _swap(c.index, axis1, axis2)
            chunk_op = op.copy().reset_key()
            out_chunk =  chunk_op.new_chunk([c], chunk_shape, index=chunk_idx)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        nsplits = _swap(in_tensor.nsplits, axis1, axis2)
        return new_op.new_tensors([in_tensor], op.outputs[0].shape,
                                  chunks=out_chunks, nsplits=nsplits)


def swapaxes(a, axis1, axis2):
    """
    Interchange two axes of a tensor.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axis1 : int
        First axis.
    axis2 : int
        Second axis.

    Returns
    -------
    a_swapped : Tensor
        If `a` is a Tensor, then a view of `a` is
        returned; otherwise a new tensor is created.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([[1,2,3]])
    >>> mt.swapaxes(x,0,1).execute()
    array([[1],
           [2],
           [3]])

    >>> x = mt.array([[[0,1],[2,3]],[[4,5],[6,7]]])
    >>> x.execute()
    array([[[0, 1],
            [2, 3]],
           [[4, 5],
            [6, 7]]])

    >>> mt.swapaxes(x,0,2).execute()
    array([[[0, 4],
            [2, 6]],
           [[1, 5],
            [3, 7]]])

    """
    axis1 = validate_axis(a.ndim, axis1)
    axis2 = validate_axis(a.ndim, axis2)

    if axis1 == axis2:
        return a

    op = TensorSwapAxes(axis1, axis2, dtype=a.dtype, sparse=a.issparse())
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Integral

import numpy as np

from .... import operands
from ...core import TENSOR_TYPE
from ..datasource import arange, array
from .core import TensorRandomOperandMixin


class TensorChoice(operands.Choice, TensorRandomOperandMixin):
    __slots__ = '_a', '_size', '_replace', '_p'

    _into_one_chunk_fields_ = ['_a', '_p']
    _input_fields_ = ['_a', '_p']

    def __init__(self, state=None, size=None, replace=None,
                 dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorChoice, self).__init__(_state=state, _size=size,
                                           _replace=replace, _dtype=dtype, _gpu=gpu, **kw)

    def __call__(self, a, p, chunk_size=None):
        return self.new_tensor([a, p], None, raw_chunk_size=chunk_size)


def choice(random_state, a, size=None, replace=True, p=None, chunk_size=None, gpu=None, **kw):
    """
    Generates a random sample from a given 1-D array

    Parameters
    -----------
    a : 1-D array-like or int
        If a tensor, a random sample is generated from its elements.
        If an int, the random sample is generated as if a were mt.arange(a)
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    replace : boolean, optional
        Whether the sample is with or without replacement
    p : 1-D array-like, optional
        The probabilities associated with each entry in a.
        If not given the sample assumes a uniform distribution over all
        entries in a.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    --------
    samples : single item or tensor
        The generated random samples

    Raises
    -------
    ValueError
        If a is an int and less than zero, if a or p are not 1-dimensional,
        if a is an array-like of size 0, if p is not a vector of
        probabilities, if a and p have different lengths, or if
        replace=False and the sample size is greater than the population
        size

    See Also
    ---------
    randint, shuffle, permutation

    Examples
    ---------
    Generate a uniform random sample from mt.arange(5) of size 3:

    >>> import mars.tensor as mt

    >>> mt.random.choice(5, 3).execute()
    array([0, 3, 4])
    >>> #This is equivalent to mt.random.randint(0,5,3)

    Generate a non-uniform random sample from np.arange(5) of size 3:

    >>> mt.random.choice(5, 3, p=[0.1, 0, 0.3, 0.6, 0]).execute()
    array([3, 3, 0])

    Generate a uniform random sample from mt.arange(5) of size 3 without
    replacement:

    >>> mt.random.choice(5, 3, replace=False).execute()
    array([3,1,0])
    >>> #This is equivalent to np.random.permutation(np.arange(5))[:3]

    Generate a non-uniform random sample from mt.arange(5) of size
    3 without replacement:

    >>> mt.random.choice(5, 3, replace=False, p=[0.1, 0, 0.3, 0.6, 0]).execute()
    array([2, 3, 0])

    Any of the above can be repeated with an arbitrary array-like
    instead of just integers. For instance:

    >>> aa_milne_arr = ['pooh', 'rabbit', 'piglet', 'Christopher']
    >>> np.random.choice(aa_milne_arr, 5, p=[0.5, 0.1, 0.1, 0.3])
    array(['pooh', 'pooh', 'pooh', 'Christopher', 'piglet'],
          dtype='|S11')
    """

    if isinstance(a, Integral):
        if a <= 0:
            raise ValueError('a must be greater than 0')
        a = arange(a, chunk_size=chunk_size)
        dtype = np.random.choice(1, size=(), p=np.array([1]) if p is not None else p).dtype
    else:
        if not isinstance(a, TENSOR_TYPE):
            a = np.asarray(a)
        a = array(a, chunk_size=a.size).rechunk(a.size)  # do rechunk if a is already a tensor
        if a.ndim != 1:
            raise ValueError('a must be one dimensional')
        dtype = a.dtype

    if p is not None:
        if not isinstance(p, TENSOR_TYPE):
            p = np.asarray(p)
            if not np.isclose(p.sum(), 1, rtol=1e-7, atol=0):
                raise ValueError('probabilities do not sum to 1')
            p = array(p, chunk_size=p.size)
        else:
            p = p.rechunk(p.size)
        if p.ndim != 1:
            raise ValueError('p must be one dimensional')

    if size is None:
        length = 1
    else:
        try:
            tuple(size)
            length = np.prod(size)
        except TypeError:
            length = size
    if replace is False and length > a.size:
        raise ValueError("Cannot take a larger sample than population when 'replace=False'")

    size = random_state._handle_size(size)
    op = TensorChoice(state=random_state._state, replace=replace,
                      size=size, dtype=dtype, gpu=gpu, **kw)
    return op(a, p, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .cov import cov


def corrcoef(x, y=None, rowvar=True):
    r"""
    Return Pearson product-moment correlation coefficients.

    Please refer to the documentation for `cov` for more detail.  The
    relationship between the correlation coefficient matrix, `R`, and the
    covariance matrix, `C`, is

    .. math:: R_{ij} = \frac{ C_{ij} } { \sqrt{ C_{ii} * C_{jj} } }

    The values of `R` are between -1 and 1, inclusive.

    Parameters
    ----------
    x : array_like
        A 1-D or 2-D array containing multiple variables and observations.
        Each row of `x` represents a variable, and each column a single
        observation of all those variables. Also see `rowvar` below.
    y : array_like, optional
        An additional set of variables and observations. `y` has the same
        shape as `x`.
    rowvar : bool, optional
        If `rowvar` is True (default), then each row represents a
        variable, with observations in the columns. Otherwise, the relationship
        is transposed: each column represents a variable, while the rows
        contain observations.

    Returns
    -------
    R : Tensor
        The correlation coefficient matrix of the variables.

    See Also
    --------
    cov : Covariance matrix

    Notes
    -----
    Due to floating point rounding the resulting tensor may not be Hermitian,
    the diagonal elements may not be 1, and the elements may not satisfy the
    inequality abs(a) <= 1. The real and imaginary parts are clipped to the
    interval [-1,  1] in an attempt to improve on that situation but is not
    much help in the complex case.

    This function accepts but discards arguments `bias` and `ddof`.  This is
    for backwards compatibility with previous versions of this function.  These
    arguments had no effect on the return values of the function and can be
    safely ignored in this and previous versions of numpy.

    """
    from ..arithmetic import sqrt
    from ..datasource import diag

    c = cov(x, y, rowvar)
    if c.ndim == 0:
        return c / c
    d = diag(c)
    d = d.reshape(d.shape[0], 1)
    sqrt_d = sqrt(d)
    return (c / sqrt_d) / sqrt_d.T



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ....operands import Nonzero
from ...core import ExecutableTuple
from ..utils import recursive_tile
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor
from .unravel_index import unravel_index


class TensorNonzero(Nonzero, TensorOperandMixin):
    def __init__(self, dtype=None, **kw):
        super(TensorNonzero, self).__init__(_dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super(TensorNonzero, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        return ExecutableTuple(self.new_tensors([a], shape=(np.nan,), output_limit=a.ndim))

    @classmethod
    def tile(cls, op):
        from ..datasource import arange

        in_tensor = op.input

        flattened = in_tensor.astype(bool).flatten()
        recursive_tile(flattened)
        indices = arange(flattened.size, dtype=np.intp, chunk_size=flattened.nsplits)
        indices = indices[flattened]
        dim_indices = unravel_index(indices, in_tensor.shape)
        [recursive_tile(ind) for ind in dim_indices]

        shapes = [t.shape for t in op.outputs]
        kws = [{'nsplits': ind.nsplits, 'chunks': ind.chunks} for ind in dim_indices]
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shapes, kws=kws, output_limit=len(shapes))


def nonzero(a):
    """
    Return the indices of the elements that are non-zero.

    Returns a tuple of tensors, one for each dimension of `a`,
    containing the indices of the non-zero elements in that
    dimension. The values in `a` are always tested and returned.
    The corresponding non-zero
    values can be obtained with::

        a[nonzero(a)]

    To group the indices by element, rather than dimension, use::

        transpose(nonzero(a))

    The result of this is always a 2-D array, with a row for
    each non-zero element.

    Parameters
    ----------
    a : array_like
        Input tensor.

    Returns
    -------
    tuple_of_arrays : tuple
        Indices of elements that are non-zero.

    See Also
    --------
    flatnonzero :
        Return indices that are non-zero in the flattened version of the input
        tensor.
    Tensor.nonzero :
        Equivalent tensor method.
    count_nonzero :
        Counts the number of non-zero elements in the input tensor.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.session import new_session

    >>> sess = new_session().as_default()

    >>> x = mt.array([[1,0,0], [0,2,0], [1,1,0]])
    >>> x.execute()
    array([[1, 0, 0],
           [0, 2, 0],
           [1, 1, 0]])
    >>> sess.run(mt.nonzero(x))
    (array([0, 1, 2, 2]), array([0, 1, 0, 1]))

    >>> x[mt.nonzero(x)].execute()  # TODO(jisheng): accomplish this after fancy indexing is supported

    >>> mt.transpose(mt.nonzero(x)).execute() # TODO(jisheng): accomplish this later

    A common use for ``nonzero`` is to find the indices of an array, where
    a condition is True.  Given an array `a`, the condition `a` > 3 is a
    boolean array and since False is interpreted as 0, np.nonzero(a > 3)
    yields the indices of the `a` where the condition is true.

    >>> a = mt.array([[1,2,3],[4,5,6],[7,8,9]])
    >>> (a > 3).execute()
    array([[False, False, False],
           [ True,  True,  True],
           [ True,  True,  True]])
    >>> sess.run(mt.nonzero(a > 3))
    (array([1, 1, 1, 2, 2, 2]), array([0, 1, 2, 0, 1, 2]))

    The ``nonzero`` method of the boolean array can also be called.

    >>> sess.run((a > 3).nonzero())
    (array([1, 1, 1, 2, 2, 2]), array([0, 1, 2, 0, 1, 2]))

    """
    a = astensor(a)
    op = TensorNonzero(dtype=np.dtype(np.intp))
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .flip import flip


def fliplr(m):
    """
    Flip tensor in the left/right direction.

    Flip the entries in each row in the left/right direction.
    Columns are preserved, but appear in a different order than before.

    Parameters
    ----------
    m : array_like
        Input tensor, must be at least 2-D.

    Returns
    -------
    f : Tensor
        A view of `m` with the columns reversed.  Since a view
        is returned, this operation is :math:`\\mathcal O(1)`.

    See Also
    --------
    flipud : Flip array in the up/down direction.
    rot90 : Rotate array counterclockwise.

    Notes
    -----
    Equivalent to m[:,::-1]. Requires the tensor to be at least 2-D.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> A = mt.diag([1.,2.,3.])
    >>> A.execute()
    array([[ 1.,  0.,  0.],
           [ 0.,  2.,  0.],
           [ 0.,  0.,  3.]])
    >>> mt.fliplr(A).execute()
    array([[ 0.,  0.,  1.],
           [ 0.,  2.,  0.],
           [ 3.,  0.,  0.]])

    >>> A = mt.random.randn(2,3,5)
    >>> mt.all(mt.fliplr(A) == A[:,::-1,...]).execute()
    True

    """
    return flip(m, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin


class TensorStandardCauchy(operands.StandardCauchy, TensorRandomOperandMixin):
    __slots__ = '_size',

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorStandardCauchy, self).__init__(_size=size, _state=state, _dtype=dtype,
                                                   _gpu=gpu, **kw)

    def __call__(self, chunk_size=None):
        return self.new_tensor(None, None, raw_chunk_size=chunk_size)


def standard_cauchy(random_state, size=None, chunk_size=None, gpu=None, **kw):
    r"""
    Draw samples from a standard Cauchy distribution with mode = 0.

    Also known as the Lorentz distribution.

    Parameters
    ----------
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    samples : Tensor or scalar
        The drawn samples.

    Notes
    -----
    The probability density function for the full Cauchy distribution is

    .. math:: P(x; x_0, \gamma) = \frac{1}{\pi \gamma \bigl[ 1+
              (\frac{x-x_0}{\gamma})^2 \bigr] }

    and the Standard Cauchy distribution just sets :math:`x_0=0` and
    :math:`\gamma=1`

    The Cauchy distribution arises in the solution to the driven harmonic
    oscillator problem, and also describes spectral line broadening. It
    also describes the distribution of values at which a line tilted at
    a random angle will cut the x axis.

    When studying hypothesis tests that assume normality, seeing how the
    tests perform on data from a Cauchy distribution is a good indicator of
    their sensitivity to a heavy-tailed distribution, since the Cauchy looks
    very much like a Gaussian distribution, but with heavier tails.

    References
    ----------
    .. [1] NIST/SEMATECH e-Handbook of Statistical Methods, "Cauchy
          Distribution",
          http://www.itl.nist.gov/div898/handbook/eda/section3/eda3663.htm
    .. [2] Weisstein, Eric W. "Cauchy Distribution." From MathWorld--A
          Wolfram Web Resource.
          http://mathworld.wolfram.com/CauchyDistribution.html
    .. [3] Wikipedia, "Cauchy distribution"
          http://en.wikipedia.org/wiki/Cauchy_distribution

    Examples
    --------
    Draw samples and plot the distribution:

    >>> import mars.tensor as mt
    >>> import matplotlib.pyplot as plt

    >>> s = mt.random.standard_cauchy(1000000)
    >>> s = s[(s>-25) & (s<25)]  # truncate distribution so it plots well
    >>> plt.hist(s.execute(), bins=100)
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().standard_cauchy(size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorStandardCauchy(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ....config import options


def tree_add(dtype, chunks, idx, shape):
    """
    Generate tree add plan.

    Assume combine size as 4, given a input chunks with size 8,
    we will generate tree add plan like:

    op op op op    op op op op
     |        |     |        |
      --------       --------
      tree_add        tree_add
          |             |
           -------------
              tree_add

    :param dtype: data type for tree added chunk
    :param chunks: input chunks
    :param idx: index of result chunk
    :param shape: shape of result chunk
    :return: result chunk
    """
    from .add import TensorTreeAdd

    combine_size = options.tensor.combine_size

    while len(chunks) > combine_size:
        new_chunks = []
        for i in range(0, len(chunks), combine_size):
            chks = chunks[i: i + combine_size]
            if len(chks) == 1:
                chk = chks[0]
            else:
                chk_op = TensorTreeAdd(dtype=dtype)
                chk = chk_op.new_chunk(chks, shape)
            new_chunks.append(chk)
        chunks = new_chunks

    op = TensorTreeAdd(dtype=dtype)
    return op.new_chunk(chunks, shape, index=idx)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .server import register_ui_handler, get_jinja_env
from ..utils import to_str
from .server import MarsWebAPI

_jinja_env = get_jinja_env()


def scheduler_list(doc, schedulers_info):
    doc.title = 'Mars UI'

    doc.template_variables['scheduler_metrics'] = schedulers_info
    doc.template = _jinja_env.get_template('scheduler_list.html')


def scheduler_detail(doc, schedulers_info, endpoint):
    doc.title = 'Mars UI'

    doc.template_variables['endpoint'] = endpoint
    doc.template_variables['scheduler_metrics'] = schedulers_info[endpoint]
    doc.template = _jinja_env.get_template('scheduler_detail.html')


def _route(scheduler_ip, doc):
    web_api = MarsWebAPI(scheduler_ip)
    schedulers_info = web_api.get_schedulers_info()

    endpoint = doc.session_context.request.arguments.get('endpoint')
    if not endpoint:
        return scheduler_list(doc, schedulers_info)
    else:
        return scheduler_detail(doc, schedulers_info, to_str(endpoint[0]))


register_ui_handler('/scheduler', _route)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin


class TensorMultinomial(operands.Multinomial, TensorRandomOperandMixin):
    __slots__ = '_n', '_pvals', '_size'

    def __init__(self, n=None, pvals=None, state=None, size=None,
                 dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorMultinomial, self).__init__(_n=n, _pvals=pvals, _state=state,
                                                _size=size, _dtype=dtype, _gpu=gpu, **kw)

    def __call__(self, chunk_size=None):
        if self._size is None:
            shape = (len(self._pvals),)
        else:
            try:
                shape = tuple(self._size) + (len(self._pvals),)
            except TypeError:
                shape = (self._size, len(self._pvals))
        return self.new_tensor(None, shape, raw_chunk_size=chunk_size)


def multinomial(random_state, n, pvals, size=None, chunk_size=None, gpu=None, **kw):
    """
    Draw samples from a multinomial distribution.

    The multinomial distribution is a multivariate generalisation of the
    binomial distribution.  Take an experiment with one of ``p``
    possible outcomes.  An example of such an experiment is throwing a dice,
    where the outcome can be 1 through 6.  Each sample drawn from the
    distribution represents `n` such experiments.  Its values,
    ``X_i = [X_0, X_1, ..., X_p]``, represent the number of times the
    outcome was ``i``.

    Parameters
    ----------
    n : int
        Number of experiments.
    pvals : sequence of floats, length p
        Probabilities of each of the ``p`` different outcomes.  These
        should sum to 1 (however, the last element is always assumed to
        account for the remaining probability, as long as
        ``sum(pvals[:-1]) <= 1)``.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor
        The drawn samples, of shape *size*, if that was provided.  If not,
        the shape is ``(N,)``.

        In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
        value drawn from the distribution.

    Examples
    --------
    Throw a dice 20 times:

    >>> import mars.tensor as mt

    >>> mt.random.multinomial(20, [1/6.]*6, size=1).execute()
    array([[4, 1, 7, 5, 2, 1]])

    It landed 4 times on 1, once on 2, etc.

    Now, throw the dice 20 times, and 20 times again:

    >>> mt.random.multinomial(20, [1/6.]*6, size=2).execute()
    array([[3, 4, 3, 3, 4, 3],
           [2, 4, 3, 4, 0, 7]])

    For the first run, we threw 3 times 1, 4 times 2, etc.  For the second,
    we threw 2 times 1, 4 times 2, etc.

    A loaded die is more likely to land on number 6:

    >>> mt.random.multinomial(100, [1/7.]*5 + [2/7.]).execute()
    array([11, 16, 14, 17, 16, 26])

    The probability inputs should be normalized. As an implementation
    detail, the value of the last entry is ignored and assumed to take
    up any leftover probability mass, but this should not be relied on.
    A biased coin which has twice as much weight on one side as on the
    other should be sampled like so:

    >>> mt.random.multinomial(100, [1.0 / 3, 2.0 / 3]).execute()  # RIGHT
    array([38, 62])

    not like:

    >>> mt.random.multinomial(100, [1.0, 2.0]).execute()  # WRONG
    array([100,   0])
    """
    n = int(n)
    pvals = tuple(pvals)
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().multinomial(n, pvals, size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorMultinomial(n=n, pvals=pvals, state=random_state._state, size=size, gpu=gpu, **kw)
    return op(chunk_size=chunk_size)




import os
import time
from collections import namedtuple

import psutil

_proc = psutil.Process()
_timer = getattr(time, 'monotonic', time.time)

if 'MARS_USE_PROCESS_STAT' in os.environ:
    _use_process_stat = bool(int(os.environ['MARS_USE_PROCESS_STAT'].strip('"')))
else:
    _use_process_stat = False

if 'MARS_CPU_TOTAL' in os.environ:
    _cpu_total = int(os.environ['MARS_CPU_TOTAL'].strip('"'))
else:
    _cpu_total = psutil.cpu_count(logical=True)

if 'MARS_MEMORY_TOTAL' in os.environ:
    _mem_total = int(os.environ['MARS_MEMORY_TOTAL'].strip('"'))
else:
    _mem_total = None

_virt_memory_stat = namedtuple('virtual_memory', 'total available percent used free')

_shm_path = [pt.mountpoint for pt in psutil.disk_partitions(all=True)
             if pt.mountpoint in ('/tmp', '/dev/shm') and pt.device == 'tmpfs']
if not _shm_path:
    _shm_path = None
else:
    _shm_path = _shm_path[0]


def virtual_memory():
    sys_mem = psutil.virtual_memory()
    if not _use_process_stat:
        total = sys_mem.total
        used = sys_mem.used + getattr(sys_mem, 'shared', 0)
        available = sys_mem.available
        free = sys_mem.free
        percent = 100.0 * (total - available) / total
        return _virt_memory_stat(total, available, percent, used, free)
    else:
        used = 0
        for p in psutil.process_iter():
            try:
                used += p.memory_info().rss
            except psutil.NoSuchProcess:
                pass

        if _shm_path:
            shm_stats = psutil.disk_usage(_shm_path)
            used += shm_stats.used

        total = min(_mem_total or sys_mem.total, sys_mem.total)
        # TODO sys_mem.available does not work in container
        # available = min(sys_mem.available, total - used)
        available = total - used
        free = min(sys_mem.free, total - used)
        percent = 100.0 * (total - available) / total
        return _virt_memory_stat(total, available, percent, used, free)


def cpu_count():
    return _cpu_total


_last_cpu_measure = None


def _take_process_cpu_snapshot():
    num_cpus = cpu_count() or 1

    def timer():
        return _timer() * num_cpus

    processes = [p for p in psutil.process_iter() if p.pid != _proc.pid]

    pts = dict()
    sts = dict()
    for p in processes:
        try:
            pts[p.pid] = p.cpu_times()
            sts[p.pid] = timer()
        except psutil.NoSuchProcess:
            pass

    pts[_proc.pid] = _proc.cpu_times()
    sts[_proc.pid] = timer()
    return pts, sts


def cpu_percent():
    global _last_cpu_measure
    if not _use_process_stat:
        return sum(psutil.cpu_percent(percpu=True))

    num_cpus = cpu_count() or 1
    pts, sts = _take_process_cpu_snapshot()

    if _last_cpu_measure is None:
        _last_cpu_measure = (pts, sts)
        return None

    old_pts, old_sts = _last_cpu_measure

    percents = []
    for pid in pts:
        if pid not in old_pts:
            continue
        pt1 = old_pts[pid]
        pt2 = pts[pid]
        delta_proc = (pt2.user - pt1.user) + (pt2.system - pt1.system)
        delta_time = sts[pid] - old_sts[pid]

        try:
            overall_cpus_percent = (delta_proc / delta_time) * 100
        except ZeroDivisionError:
            percents.append(0.0)
        else:
            single_cpu_percent = overall_cpus_percent * num_cpus
            percents.append(single_cpu_percent)
    _last_cpu_measure = (pts, sts)
    return round(sum(percents), 1)


def disk_usage(d):
    return psutil.disk_usage(d)


def iowait():
    cpu_percent = psutil.cpu_times_percent()
    try:
        return cpu_percent.iowait
    except AttributeError:
        return None


_last_disk_io_meta = None


def disk_io_usage():
    global _last_disk_io_meta

    disk_counters = psutil.disk_io_counters()
    tst = time.time()

    read_bytes = disk_counters.read_bytes
    write_bytes = disk_counters.write_bytes
    if _last_disk_io_meta is None:
        _last_disk_io_meta = (read_bytes, write_bytes, tst)
        return None

    last_read_bytes, last_write_bytes, last_time = _last_disk_io_meta
    delta_time = tst - last_time
    read_speed = (read_bytes - last_read_bytes) / delta_time
    write_speed = (write_bytes - last_write_bytes) / delta_time

    _last_disk_io_meta = (read_bytes, write_bytes, tst)
    return read_speed, write_speed


_last_net_io_meta = None


def net_io_usage():
    global _last_net_io_meta

    net_counters = psutil.net_io_counters()
    tst = time.time()

    send_bytes = net_counters.bytes_sent
    recv_bytes = net_counters.bytes_recv
    if _last_net_io_meta is None:
        _last_net_io_meta = (send_bytes, recv_bytes, tst)
        return None

    last_send_bytes, last_recv_bytes, last_time = _last_net_io_meta
    delta_time = tst - last_time
    recv_speed = (recv_bytes - last_recv_bytes) / delta_time
    send_speed = (send_bytes - last_send_bytes) / delta_time

    _last_net_io_meta = (send_bytes, recv_bytes, tst)
    return recv_speed, send_speed



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import time
import logging

import requests

from ..compat import six, TimeoutError
from ..serialize import dataserializer
from ..errors import ResponseMalformed, ExecutionInterrupted, ExecutionFailed, \
    ExecutionStateUnknown, ExecutionNotStopped
from ..graph import DirectedGraph

logger = logging.getLogger(__name__)


class Session(object):
    def __init__(self, endpoint, args=None):
        self._endpoint = endpoint
        self._args = args
        self._tensor_to_graph = dict()

        self._req_session = requests.Session()

        from requests.adapters import HTTPAdapter
        self._req_session.mount('http://stackoverflow.com', HTTPAdapter(max_retries=5))
        self._main()

    @property
    def session_id(self):
        return self._session_id

    @property
    def endpoint(self):
        return self._endpoint

    @endpoint.setter
    def endpoint(self, url):
        self._endpoint = url

    def _main(self):
        resp = self._req_session.post(self._endpoint + '/api/session', self._args)
        if resp.status_code >= 400:
            raise SystemError('Failed to create mars session.')
        content = json.loads(resp.text)
        self._session_id = content['session_id']

    def _check_response_finished(self, graph_url):
        try:
            resp = self._req_session.get(graph_url)
        except requests.ConnectionError as ex:
            err_msg = str(ex)
            if 'ConnectionResetError' in err_msg or 'Connection refused' in err_msg:
                return False
            raise

        if resp.status_code == 504:
            logging.debug('Gateway Time-out, try again')
            return False
        if resp.status_code >= 400:
            raise SystemError('Failed to obtain execution status. Code: %d, Reason: %s, Content:\n%s' %
                              (resp.status_code, resp.reason, resp.text))
        try:
            resp_json = json.loads(resp.text)
        except ValueError:
            raise ResponseMalformed('Response malformed. Code: %d, Content:\n%s' %
                                    (resp.status_code, resp.text))
        if resp_json['state'] == 'success':
            return True
        elif resp_json['state'] in ('running', 'preparing'):
            return False
        elif resp_json['state'] in ('cancelled', 'cancelling'):
            raise ExecutionInterrupted
        elif resp_json['state'] == 'failed':
            # TODO add traceback
            if 'traceback' in resp_json:
                traceback = resp_json['traceback']
                traceback = ''.join(str(s) for s in traceback) \
                    if isinstance(traceback, list) else traceback
                raise ExecutionFailed(
                    'Graph execution failed.\nMessage: %s\nTraceback from server:\n%s' %
                    (resp_json['msg'], traceback))
            else:
                raise ExecutionFailed('Graph execution failed with unknown reason.')
        raise ExecutionStateUnknown(
            'Unknown graph execution state %s' % resp_json['state'])

    def run(self, *tensors, **kw):
        timeout = kw.pop('timeout', -1)
        compose = kw.pop('compose', True)
        wait = kw.pop('wait', True)
        if kw:
            raise TypeError('run got unexpected key arguments {0}'.format(', '.join(kw.keys())))

        graph = DirectedGraph()
        for t in tensors:
            graph = t.build_graph(graph=graph, tiled=False, compose=compose)
        targets = [t.key for t in tensors]

        targets_join = ','.join(targets)
        session_url = self._endpoint + '/api/session/' + self._session_id
        graph_json = graph.to_json()

        resp_json = self._submit_graph(graph_json, targets_join)
        graph_key = resp_json['graph_key']
        graph_url = session_url + '/graph/' + graph_key

        for t in tensors:
            self._tensor_to_graph[t.key] = graph_key

        if not wait:
            return graph_key

        exec_start_time = time.time()
        while timeout <= 0 or time.time() - exec_start_time <= timeout:
            try:
                time.sleep(1)
                if self._check_response_finished(graph_url):
                    break
            except KeyboardInterrupt:
                resp = self._req_session.delete(graph_url)
                if resp.status_code >= 400:
                    raise ExecutionNotStopped(
                        'Failed to stop graph execution. Code: %d, Reason: %s, Content:\n%s' %
                        (resp.status_code, resp.reason, resp.text))
        if 0 < timeout < time.time() - exec_start_time:
            raise TimeoutError
        data_list = []
        for tk in targets:
            resp = self._req_session.get(session_url + '/graph/' + graph_key + '/data/' + tk)
            if resp.status_code >= 400:
                continue
            data_list.append(dataserializer.loads(resp.content))
        return data_list

    def fetch(self, tensor):
        key = tensor.key
        session_url = self._endpoint + '/api/session/' + self._session_id
        data_url = session_url + '/graph/%s/data/%s' % (self._tensor_to_graph[key], key)
        resp = self._req_session.get(data_url)
        if resp.status_code >= 400:
            raise ValueError('Failed to fetch data from server. Code: %d, Reason: %s, Content:\n%s' %
                             (resp.status_code, resp.reason, resp.text))
        return dataserializer.loads(resp.content)

    def decref(self, *keys):
        session_url = self._endpoint + '/api/session/' + self._session_id
        for k in keys:
            if k not in self._tensor_to_graph:
                continue
            data_url = session_url + '/graph/%s/data/%s' % (self._tensor_to_graph[k], k)
            self._req_session.delete(data_url)

    def stop(self, graph_key):
        session_url = self._endpoint + '/api/session/' + self._session_id
        graph_url = session_url + '/graph/' + graph_key
        resp = self._req_session.delete(graph_url)
        if resp.status_code >= 400:
            raise SystemError('Failed to stop graph execution. Code: %d, Reason: %s, Content:\n%s' %
                              (resp.status_code, resp.reason, resp.text))

    def _submit_graph(self, graph_json, targets):
        session_url = self._endpoint + '/api/session/' + self._session_id
        resp = self._req_session.post(session_url + '/graph', dict(
            graph=json.dumps(graph_json),
            target=targets,
        ))
        if resp.status_code >= 400:
            resp_json = json.loads(resp.text)
            exc_info = base64.b64decode(resp_json['exc_info'])
            six.reraise(*exc_info)
        resp_json = json.loads(resp.text)
        return resp_json

    def close(self):
        self.decref(*list(self._tensor_to_graph.keys()))

        resp = self._req_session.delete(self._endpoint + '/api/session/' + self._session_id)
        if resp.status_code >= 400:
            raise SystemError('Failed to close mars session.')

    def check_service_ready(self, timeout=1):
        try:
            resp = self._req_session.get(self._endpoint + '/api', timeout=timeout)
        except (requests.ConnectionError, requests.Timeout):
            return False
        if resp.status_code >= 400:
            return False
        return True

    def count_workers(self):
        resp = self._req_session.get(self._endpoint + '/api/worker', timeout=1)
        return json.loads(resp.text)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/operand.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mars.serialize.protos import value_pb2 as mars_dot_serialize_dot_protos_dot_value__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/operand.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n#mars/serialize/protos/operand.proto\x1a!mars/serialize/protos/value.proto\"\x83\'\n\nOperandDef\x12\x0c\n\x04type\x18\x01 \x01(\t\x12#\n\x04\x61ttr\x18\x02 \x03(\x0b\x32\x15.OperandDef.AttrEntry\x12\x1d\n\x08operands\x18\x03 \x03(\x0b\x32\x0b.OperandDef\x12\n\n\x02id\x18\x04 \x01(\t\x1a\x33\n\tAttrEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x15\n\x05value\x18\x02 \x01(\x0b\x32\x06.Value:\x02\x38\x01\"\xe1%\n\x0bOperandType\x12\x08\n\x04NULL\x10\x00\x12\n\n\x06SCALAR\x10\x01\x12\x16\n\x12TENSOR_DATA_SOURCE\x10\x02\x12\x0f\n\x0bTENSOR_ONES\x10\x03\x12\x14\n\x10TENSOR_ONES_LIKE\x10\x04\x12\x10\n\x0cTENSOR_ZEROS\x10\x05\x12\x15\n\x11TENSOR_ZEROS_LIKE\x10\x06\x12\x10\n\x0cTENSOR_EMPTY\x10\x07\x12\x15\n\x11TENSOR_EMPTY_LIKE\x10\x08\x12\x0f\n\x0bTENSOR_FULL\x10\t\x12\x11\n\rTENSOR_ARANGE\x10\n\x12\x12\n\x0eTENSOR_INDICES\x10\x0b\x12\x0f\n\x0bTENSOR_DIAG\x10\x0c\x12\x0e\n\nTENSOR_EYE\x10\r\x12\x13\n\x0fTENSOR_LINSPACE\x10\x0e\x12\x0f\n\x0bTENSOR_TRIU\x10\x0f\x12\x0f\n\x0bTENSOR_TRIL\x10\x10\x12\r\n\tRAND_RAND\x10)\x12\x0e\n\nRAND_RANDN\x10*\x12\x10\n\x0cRAND_RANDINT\x10+\x12\x18\n\x14RAND_RANDOM_INTEGERS\x10,\x12\x16\n\x12RAND_RANDOM_SAMPLE\x10-\x12\x0f\n\x0bRAND_RANDOM\x10.\x12\r\n\tRAND_RANF\x10/\x12\x0f\n\x0bRAND_SAMPLE\x10\x30\x12\x0e\n\nRAND_BYTES\x10\x31\x12\r\n\tRAND_BETA\x10\x32\x12\x11\n\rRAND_BINOMIAL\x10\x33\x12\x12\n\x0eRAND_CHISQUARE\x10\x34\x12\x0f\n\x0bRAND_CHOICE\x10\x35\x12\x12\n\x0eRAND_DIRICHLET\x10\x36\x12\x14\n\x10RAND_EXPONENTIAL\x10\x37\x12\n\n\x06RAND_F\x10\x38\x12\x0e\n\nRAND_GAMMA\x10\x39\x12\x12\n\x0eRAND_GEOMETRIC\x10:\x12\x0f\n\x0bRAND_GUMBEL\x10;\x12\x17\n\x13RAND_HYPERGEOMETRIC\x10<\x12\x10\n\x0cRAND_LAPLACE\x10=\x12\x11\n\rRAND_LOGISTIC\x10>\x12\x12\n\x0eRAND_LOGNORMAL\x10?\x12\x12\n\x0eRAND_LOGSERIES\x10@\x12\x14\n\x10RAND_MULTINOMIAL\x10\x41\x12\x1c\n\x18RAND_MULTIVARIATE_NORMAL\x10\x42\x12\x1a\n\x16RAND_NEGATIVE_BINOMIAL\x10\x43\x12\x1c\n\x18RAND_NONCENTRAL_CHISQURE\x10\x44\x12\x15\n\x11RAND_NONCENTRAL_F\x10\x45\x12\x0f\n\x0bRAND_NORMAL\x10\x46\x12\x0f\n\x0bRAND_PARETO\x10G\x12\x14\n\x10RAND_PERMUTATION\x10H\x12\x10\n\x0cRAND_POSSION\x10I\x12\x0e\n\nRAND_POWER\x10J\x12\x11\n\rRAND_RAYLEIGH\x10K\x12\x10\n\x0cRAND_SHUFFLE\x10L\x12\x18\n\x14RAND_STANDARD_CAUCHY\x10M\x12\x1d\n\x19RAND_STANDARD_EXPONENTIAL\x10N\x12\x18\n\x14RAND_STANDARD_GAMMMA\x10O\x12\x18\n\x14RAND_STANDARD_NORMAL\x10P\x12\x13\n\x0fRAND_STANDARD_T\x10Q\x12\x11\n\rRAND_TOMAXINT\x10R\x12\x13\n\x0fRAND_TRIANGULAR\x10S\x12\x10\n\x0cRAND_UNIFORM\x10T\x12\x11\n\rRAND_VONMISES\x10U\x12\r\n\tRAND_WALD\x10V\x12\x10\n\x0cRAND_WEIBULL\x10W\x12\r\n\tRAND_ZIPF\x10X\x12\x07\n\x03\x41\x44\x44\x10\x65\x12\x10\n\x0c\x41\x44\x44_CONSTANT\x10\x66\x12\x07\n\x03SUB\x10g\x12\x10\n\x0cSUB_CONSTANT\x10h\x12\x07\n\x03MUL\x10i\x12\x10\n\x0cMUL_CONSTANT\x10j\x12\x07\n\x03\x44IV\x10k\x12\x10\n\x0c\x44IV_CONSTANT\x10l\x12\x0b\n\x07TRUEDIV\x10m\x12\x11\n\rTDIV_CONSTANT\x10n\x12\x0c\n\x08\x46LOORDIV\x10o\x12\x11\n\rFDIV_CONSTANT\x10p\x12\x07\n\x03POW\x10q\x12\x10\n\x0cPOW_CONSTANT\x10r\x12\x07\n\x03MOD\x10s\x12\x10\n\x0cMOD_CONSTANT\x10t\x12\x08\n\x04\x46MOD\x10u\x12\x11\n\rFMOD_CONSTANT\x10v\x12\r\n\tLOGADDEXP\x10w\x12\x10\n\x0cLAE_CONSTANT\x10x\x12\x0e\n\nLOGADDEXP2\x10y\x12\x11\n\rLAE2_CONSTANT\x10z\x12\x0c\n\x08NEGATIVE\x10{\x12\x0c\n\x08POSITIVE\x10|\x12\x0c\n\x08\x41\x42SOLUTE\x10}\x12\x08\n\x04\x46\x41\x42S\x10~\x12\x07\n\x03\x41\x42S\x10\x7f\x12\t\n\x04RINT\x10\x80\x01\x12\t\n\x04SIGN\x10\x81\x01\x12\t\n\x04\x43ONJ\x10\x82\x01\x12\x08\n\x03\x45XP\x10\x83\x01\x12\t\n\x04\x45XP2\x10\x84\x01\x12\x08\n\x03LOG\x10\x85\x01\x12\t\n\x04LOG2\x10\x86\x01\x12\n\n\x05LOG10\x10\x87\x01\x12\n\n\x05\x45XPM1\x10\x88\x01\x12\n\n\x05LOG1P\x10\x89\x01\x12\t\n\x04SQRT\x10\x8a\x01\x12\x0b\n\x06SQUARE\x10\x8b\x01\x12\t\n\x04\x43\x42RT\x10\x8c\x01\x12\x0f\n\nRECIPROCAL\x10\x8d\x01\x12\x07\n\x02\x45Q\x10\x8e\x01\x12\x10\n\x0b\x45Q_CONSTANT\x10\x8f\x01\x12\x07\n\x02NE\x10\x90\x01\x12\x10\n\x0bNE_CONSTANT\x10\x91\x01\x12\x07\n\x02LT\x10\x92\x01\x12\x10\n\x0bLT_CONSTANT\x10\x93\x01\x12\x07\n\x02LE\x10\x94\x01\x12\x10\n\x0bLE_CONSTANT\x10\x95\x01\x12\x07\n\x02GT\x10\x96\x01\x12\x10\n\x0bGT_CONSTANT\x10\x97\x01\x12\x07\n\x02GE\x10\x98\x01\x12\x10\n\x0bGE_CONSTANT\x10\x99\x01\x12\x08\n\x03SIN\x10\x9a\x01\x12\x08\n\x03\x43OS\x10\x9b\x01\x12\x08\n\x03TAN\x10\x9c\x01\x12\x0b\n\x06\x41RCSIN\x10\x9d\x01\x12\x0b\n\x06\x41RCCOS\x10\x9e\x01\x12\x0b\n\x06\x41RCTAN\x10\x9f\x01\x12\x0c\n\x07\x41RCTAN2\x10\xa0\x01\x12\x13\n\x0e\x41RCT2_CONSTANT\x10\xa1\x01\x12\n\n\x05HYPOT\x10\xa2\x01\x12\x13\n\x0eHYPOT_CONSTANT\x10\xa3\x01\x12\t\n\x04SINH\x10\xa4\x01\x12\t\n\x04\x43OSH\x10\xa5\x01\x12\t\n\x04TANH\x10\xa6\x01\x12\x0c\n\x07\x41RCSINH\x10\xa7\x01\x12\x0c\n\x07\x41RCCOSH\x10\xa8\x01\x12\x0c\n\x07\x41RCTANH\x10\xa9\x01\x12\x0c\n\x07\x44\x45G2RAD\x10\xaa\x01\x12\x0c\n\x07RAD2DEG\x10\xab\x01\x12\x0b\n\x06\x42ITAND\x10\xac\x01\x12\x14\n\x0f\x42ITAND_CONSTANT\x10\xad\x01\x12\n\n\x05\x42ITOR\x10\xae\x01\x12\x13\n\x0e\x42ITOR_CONSTANT\x10\xaf\x01\x12\x0b\n\x06\x42ITXOR\x10\xb0\x01\x12\x14\n\x0f\x42ITXOR_CONSTANT\x10\xb1\x01\x12\x0b\n\x06INVERT\x10\xb2\x01\x12\x0b\n\x06LSHIFT\x10\xb3\x01\x12\x14\n\x0fLSHIFT_CONSTANT\x10\xb4\x01\x12\x0b\n\x06RSHIFT\x10\xb5\x01\x12\x14\n\x0fRSHIFT_CONSTANT\x10\xb6\x01\x12\x08\n\x03\x41ND\x10\xb7\x01\x12\x11\n\x0c\x41ND_CONSTANT\x10\xb8\x01\x12\x07\n\x02OR\x10\xb9\x01\x12\x10\n\x0bOR_CONSTANT\x10\xba\x01\x12\x08\n\x03XOR\x10\xbb\x01\x12\x11\n\x0cXOR_CONSTANT\x10\xbc\x01\x12\x08\n\x03NOT\x10\xbd\x01\x12\x0c\n\x07MAXIMUM\x10\xbe\x01\x12\x15\n\x10MAXIMUM_CONSTANT\x10\xbf\x01\x12\x0c\n\x07MINIMUM\x10\xc0\x01\x12\x15\n\x10MINIMUM_CONSTANT\x10\xc1\x01\x12\x0b\n\x06\x41ROUND\x10\xc2\x01\x12\x10\n\x0b\x46LOAT_POWER\x10\xc3\x01\x12\x19\n\x14\x46LOAT_POWER_CONSTANT\x10\xc4\x01\x12\t\n\x04\x46MAX\x10\xc5\x01\x12\x12\n\rFMAX_CONSTANT\x10\xc6\x01\x12\t\n\x04\x46MIN\x10\xc7\x01\x12\x12\n\rFMIN_CONSTANT\x10\xc8\x01\x12\r\n\x08ISFINITE\x10\xc9\x01\x12\n\n\x05ISINF\x10\xca\x01\x12\n\n\x05ISNAN\x10\xcb\x01\x12\x0c\n\x07SIGNBIT\x10\xcc\x01\x12\r\n\x08\x43OPYSIGN\x10\xcd\x01\x12\x16\n\x11\x43OPYSIGN_CONSTANT\x10\xce\x01\x12\x0e\n\tNEXTAFTER\x10\xcf\x01\x12\x17\n\x12NEXTAFTER_CONSTANT\x10\xd0\x01\x12\x0c\n\x07SPACING\x10\xd1\x01\x12\n\n\x05LDEXP\x10\xd2\x01\x12\x13\n\x0eLDEXP_CONSTANT\x10\xd3\x01\x12\n\n\x05\x46REXP\x10\xd4\x01\x12\t\n\x04MODF\x10\xd5\x01\x12\n\n\x05\x46LOOR\x10\xd6\x01\x12\t\n\x04\x43\x45IL\x10\xd7\x01\x12\n\n\x05TRUNC\x10\xd8\x01\x12\x0c\n\x07\x44\x45GREES\x10\xd9\x01\x12\x0c\n\x07RADIANS\x10\xda\x01\x12\t\n\x04\x43LIP\x10\xdb\x01\x12\x0b\n\x06ISREAL\x10\xdc\x01\x12\x0e\n\tISCOMPLEX\x10\xdd\x01\x12\t\n\x04REAL\x10\xde\x01\x12\t\n\x04IMAG\x10\xdf\x01\x12\x08\n\x03\x46IX\x10\xe0\x01\x12\x07\n\x02I0\x10\xe1\x01\x12\t\n\x04SINC\x10\xe2\x01\x12\x0f\n\nNAN_TO_NUM\x10\xe3\x01\x12\x0c\n\x07ISCLOSE\x10\xe4\x01\x12\x15\n\x10ISCLOSE_CONSTANT\x10\xe5\x01\x12\x0b\n\x06\x44IVMOD\x10\xe7\x01\x12\n\n\x05\x41NGLE\x10\xe8\x01\x12\r\n\x08SET_REAL\x10\xe9\x01\x12\x16\n\x11SET_REAL_CONSTANT\x10\xea\x01\x12\r\n\x08SET_IMAG\x10\xeb\x01\x12\x16\n\x11SET_IMAG_CONSTANT\x10\xec\x01\x12\r\n\x08TREE_ADD\x10\xfb\x01\x12\x12\n\rTREE_MULTIPLY\x10\xfc\x01\x12\x0b\n\x06\x43UMSUM\x10\xad\x02\x12\x0c\n\x07\x43UMPROD\x10\xae\x02\x12\t\n\x04PROD\x10\xaf\x02\x12\x08\n\x03SUM\x10\xb0\x02\x12\x08\n\x03MAX\x10\xb1\x02\x12\x08\n\x03MIN\x10\xb2\x02\x12\x08\n\x03\x41LL\x10\xb3\x02\x12\x08\n\x03\x41NY\x10\xb4\x02\x12\x0f\n\nMEAN_CHUNK\x10\xb5\x02\x12\x11\n\x0cMEAN_COMBINE\x10\xb6\x02\x12\t\n\x04MEAN\x10\xb7\x02\x12\x0b\n\x06\x41RGMAX\x10\xb8\x02\x12\x11\n\x0c\x41RGMAX_CHUNK\x10\xb9\x02\x12\x13\n\x0e\x41RGMAX_COMBINE\x10\xba\x02\x12\x0b\n\x06\x41RGMIN\x10\xbb\x02\x12\x11\n\x0c\x41RGMIN_CHUNK\x10\xbc\x02\x12\x13\n\x0e\x41RGMIN_COMBINE\x10\xbd\x02\x12\x0b\n\x06NANSUM\x10\xbe\x02\x12\x0b\n\x06NANMAX\x10\xbf\x02\x12\x0b\n\x06NANMIN\x10\xc0\x02\x12\x0c\n\x07NANPROD\x10\xc1\x02\x12\x0c\n\x07NANMEAN\x10\xc2\x02\x12\x12\n\rNANMEAN_CHUNK\x10\xc3\x02\x12\x0e\n\tNANARGMAX\x10\xc4\x02\x12\x14\n\x0fNANARGMAX_CHUNK\x10\xc5\x02\x12\x16\n\x11NANARGMAX_COMBINE\x10\xc6\x02\x12\x0e\n\tNANARGMIN\x10\xc7\x02\x12\x14\n\x0fNANARGMIN_CHUNK\x10\xc8\x02\x12\x16\n\x11NANARGMIN_COMBINE\x10\xc9\x02\x12\x12\n\rCOUNT_NONZERO\x10\xca\x02\x12\x11\n\x0cMOMENT_CHUNK\x10\xcb\x02\x12\x14\n\x0fNANMOMENT_CHUNK\x10\xcc\x02\x12\x13\n\x0eMOMENT_COMBINE\x10\xcd\x02\x12\x16\n\x11NANMOMENT_COMBINE\x10\xce\x02\x12\x0b\n\x06MOMENT\x10\xcf\x02\x12\x0e\n\tNANMOMENT\x10\xd0\x02\x12\x08\n\x03VAR\x10\xd1\x02\x12\x08\n\x03STD\x10\xd2\x02\x12\x0b\n\x06NANVAR\x10\xd3\x02\x12\x0b\n\x06NANSTD\x10\xd4\x02\x12\x0e\n\tNANCUMSUM\x10\xd5\x02\x12\x0f\n\nNANCUMPROD\x10\xd6\x02\x12\x0c\n\x07RESHAPE\x10\x91\x03\x12\n\n\x05SLICE\x10\x92\x03\x12\n\n\x05INDEX\x10\x93\x03\x12\x12\n\rINDEXSETVALUE\x10\x94\x03\x12\x10\n\x0b\x43ONCATENATE\x10\x95\x03\x12\x0c\n\x07RECHUNK\x10\x96\x03\x12\x0b\n\x06\x41STYPE\x10\x97\x03\x12\x0e\n\tTRANSPOSE\x10\x98\x03\x12\r\n\x08SWAPAXES\x10\x99\x03\x12\x11\n\x0c\x42ROADCAST_TO\x10\x9a\x03\x12\n\n\x05STACK\x10\x9b\x03\x12\n\n\x05WHERE\x10\x9c\x03\x12\x0b\n\x06\x43HOOSE\x10\x9d\x03\x12\x0c\n\x07NONZERO\x10\x9e\x03\x12\r\n\x08\x41RGWHERE\x10\x9f\x03\x12\x12\n\rUNRAVEL_INDEX\x10\xa0\x03\x12\x16\n\x11RAVEL_MULTI_INDEX\x10\xa1\x03\x12\x10\n\x0b\x41RRAY_SPLIT\x10\xa2\x03\x12\x0c\n\x07SQUEEZE\x10\xa3\x03\x12\r\n\x08\x44IGITIZE\x10\xa4\x03\x12\x0b\n\x06REPEAT\x10\xa5\x03\x12\x0b\n\x06\x43OPYTO\x10\xa6\x03\x12\t\n\x04ISIN\x10\xa7\x03\x12\x0e\n\tTENSORDOT\x10\xf5\x03\x12\x08\n\x03\x44OT\x10\xf6\x03\x12\x0b\n\x06MATMUL\x10\xf7\x03\x12\r\n\x08\x43HOLESKY\x10\xfe\x03\x12\x07\n\x02QR\x10\xff\x03\x12\x08\n\x03SVD\x10\x80\x04\x12\x07\n\x02LU\x10\x81\x04\x12\x15\n\x10SOLVE_TRIANGULAR\x10\x88\x04\x12\t\n\x04NORM\x10\x92\x04\x12\x08\n\x03\x46\x46T\x10\xd9\x04\x12\t\n\x04IFFT\x10\xda\x04\x12\t\n\x04\x46\x46T2\x10\xdb\x04\x12\n\n\x05IFFT2\x10\xdc\x04\x12\t\n\x04\x46\x46TN\x10\xdd\x04\x12\n\n\x05IFFTN\x10\xde\x04\x12\t\n\x04RFFT\x10\xdf\x04\x12\n\n\x05IRFFT\x10\xe0\x04\x12\n\n\x05RFFT2\x10\xe1\x04\x12\x0b\n\x06IRFFT2\x10\xe2\x04\x12\n\n\x05RFFTN\x10\xe3\x04\x12\x0b\n\x06IRFFTN\x10\xe4\x04\x12\t\n\x04HFFT\x10\xe5\x04\x12\n\n\x05IHFFT\x10\xe6\x04\x12\x0c\n\x07\x46\x46TFREQ\x10\xe7\x04\x12\x12\n\rFFTFREQ_CHUNK\x10\xe8\x04\x12\r\n\x08RFFTFREQ\x10\xe9\x04\x12\r\n\x08\x46\x46TSHIFT\x10\xea\x04\x12\x0e\n\tIFFTSHIFT\x10\xeb\x04\x12\x1e\n\x19SPARSE_MATRIX_DATA_SOURCE\x10\xbd\x05\x12\x14\n\x0f\x44\x45NSE_TO_SPARSE\x10\xbe\x05\x12\x14\n\x0fSPARSE_TO_DENSE\x10\xbf\x05\x12\t\n\x04\x46USE\x10\xa1\x06\x12\n\n\x05\x45NTER\x10\x85\x07\x12\n\n\x05LEAVE\x10\x86\x07\x12\x0f\n\nFIX_LATEST\x10\x87\x07\x12\x0c\n\x07IF_ELSE\x10\x88\x07\x12\x0e\n\tNEXT_ITER\x10\x89\x07\x12\x0e\n\tTABLE_COO\x10\xeb\x07\x12\x0e\n\tSTORE_COO\x10\xec\x07\x12\x0b\n\x05\x46\x45TCH\x10\xbf\x84=b\x06proto3')
  ,
  dependencies=[mars_dot_serialize_dot_protos_dot_value__pb2.DESCRIPTOR,])



_OPERANDDEF_OPERANDTYPE = _descriptor.EnumDescriptor(
  name='OperandType',
  full_name='OperandDef.OperandType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NULL', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SCALAR', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_DATA_SOURCE', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ONES', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ONES_LIKE', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ZEROS', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ZEROS_LIKE', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_EMPTY', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_EMPTY_LIKE', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FULL', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ARANGE', index=10, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_INDICES', index=11, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_DIAG', index=12, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_EYE', index=13, number=13,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_LINSPACE', index=14, number=14,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_TRIU', index=15, number=15,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_TRIL', index=16, number=16,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RAND', index=17, number=41,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDN', index=18, number=42,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDINT', index=19, number=43,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDOM_INTEGERS', index=20, number=44,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDOM_SAMPLE', index=21, number=45,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDOM', index=22, number=46,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANF', index=23, number=47,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_SAMPLE', index=24, number=48,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_BYTES', index=25, number=49,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_BETA', index=26, number=50,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_BINOMIAL', index=27, number=51,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_CHISQUARE', index=28, number=52,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_CHOICE', index=29, number=53,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_DIRICHLET', index=30, number=54,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_EXPONENTIAL', index=31, number=55,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_F', index=32, number=56,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_GAMMA', index=33, number=57,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_GEOMETRIC', index=34, number=58,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_GUMBEL', index=35, number=59,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_HYPERGEOMETRIC', index=36, number=60,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LAPLACE', index=37, number=61,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LOGISTIC', index=38, number=62,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LOGNORMAL', index=39, number=63,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LOGSERIES', index=40, number=64,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_MULTINOMIAL', index=41, number=65,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_MULTIVARIATE_NORMAL', index=42, number=66,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NEGATIVE_BINOMIAL', index=43, number=67,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NONCENTRAL_CHISQURE', index=44, number=68,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NONCENTRAL_F', index=45, number=69,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NORMAL', index=46, number=70,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_PARETO', index=47, number=71,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_PERMUTATION', index=48, number=72,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_POSSION', index=49, number=73,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_POWER', index=50, number=74,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RAYLEIGH', index=51, number=75,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_SHUFFLE', index=52, number=76,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_CAUCHY', index=53, number=77,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_EXPONENTIAL', index=54, number=78,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_GAMMMA', index=55, number=79,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_NORMAL', index=56, number=80,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_T', index=57, number=81,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_TOMAXINT', index=58, number=82,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_TRIANGULAR', index=59, number=83,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_UNIFORM', index=60, number=84,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_VONMISES', index=61, number=85,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_WALD', index=62, number=86,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_WEIBULL', index=63, number=87,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_ZIPF', index=64, number=88,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ADD', index=65, number=101,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ADD_CONSTANT', index=66, number=102,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUB', index=67, number=103,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUB_CONSTANT', index=68, number=104,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MUL', index=69, number=105,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MUL_CONSTANT', index=70, number=106,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIV', index=71, number=107,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIV_CONSTANT', index=72, number=108,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRUEDIV', index=73, number=109,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TDIV_CONSTANT', index=74, number=110,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOORDIV', index=75, number=111,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FDIV_CONSTANT', index=76, number=112,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POW', index=77, number=113,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POW_CONSTANT', index=78, number=114,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOD', index=79, number=115,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOD_CONSTANT', index=80, number=116,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMOD', index=81, number=117,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMOD_CONSTANT', index=82, number=118,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOGADDEXP', index=83, number=119,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LAE_CONSTANT', index=84, number=120,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOGADDEXP2', index=85, number=121,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LAE2_CONSTANT', index=86, number=122,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEGATIVE', index=87, number=123,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POSITIVE', index=88, number=124,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ABSOLUTE', index=89, number=125,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FABS', index=90, number=126,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ABS', index=91, number=127,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RINT', index=92, number=128,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIGN', index=93, number=129,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONJ', index=94, number=130,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXP', index=95, number=131,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXP2', index=96, number=132,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG', index=97, number=133,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG2', index=98, number=134,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG10', index=99, number=135,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXPM1', index=100, number=136,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG1P', index=101, number=137,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQRT', index=102, number=138,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQUARE', index=103, number=139,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CBRT', index=104, number=140,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RECIPROCAL', index=105, number=141,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EQ', index=106, number=142,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EQ_CONSTANT', index=107, number=143,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NE', index=108, number=144,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NE_CONSTANT', index=109, number=145,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LT', index=110, number=146,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LT_CONSTANT', index=111, number=147,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LE', index=112, number=148,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LE_CONSTANT', index=113, number=149,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GT', index=114, number=150,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GT_CONSTANT', index=115, number=151,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GE', index=116, number=152,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GE_CONSTANT', index=117, number=153,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIN', index=118, number=154,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COS', index=119, number=155,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TAN', index=120, number=156,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCSIN', index=121, number=157,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCCOS', index=122, number=158,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCTAN', index=123, number=159,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCTAN2', index=124, number=160,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCT2_CONSTANT', index=125, number=161,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HYPOT', index=126, number=162,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HYPOT_CONSTANT', index=127, number=163,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SINH', index=128, number=164,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COSH', index=129, number=165,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TANH', index=130, number=166,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCSINH', index=131, number=167,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCCOSH', index=132, number=168,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCTANH', index=133, number=169,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEG2RAD', index=134, number=170,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAD2DEG', index=135, number=171,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITAND', index=136, number=172,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITAND_CONSTANT', index=137, number=173,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITOR', index=138, number=174,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITOR_CONSTANT', index=139, number=175,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITXOR', index=140, number=176,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITXOR_CONSTANT', index=141, number=177,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INVERT', index=142, number=178,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LSHIFT', index=143, number=179,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LSHIFT_CONSTANT', index=144, number=180,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RSHIFT', index=145, number=181,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RSHIFT_CONSTANT', index=146, number=182,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AND', index=147, number=183,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AND_CONSTANT', index=148, number=184,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OR', index=149, number=185,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OR_CONSTANT', index=150, number=186,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='XOR', index=151, number=187,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='XOR_CONSTANT', index=152, number=188,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NOT', index=153, number=189,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAXIMUM', index=154, number=190,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAXIMUM_CONSTANT', index=155, number=191,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MINIMUM', index=156, number=192,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MINIMUM_CONSTANT', index=157, number=193,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AROUND', index=158, number=194,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT_POWER', index=159, number=195,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT_POWER_CONSTANT', index=160, number=196,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMAX', index=161, number=197,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMAX_CONSTANT', index=162, number=198,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMIN', index=163, number=199,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMIN_CONSTANT', index=164, number=200,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISFINITE', index=165, number=201,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISINF', index=166, number=202,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISNAN', index=167, number=203,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIGNBIT', index=168, number=204,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COPYSIGN', index=169, number=205,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COPYSIGN_CONSTANT', index=170, number=206,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEXTAFTER', index=171, number=207,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEXTAFTER_CONSTANT', index=172, number=208,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPACING', index=173, number=209,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LDEXP', index=174, number=210,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LDEXP_CONSTANT', index=175, number=211,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FREXP', index=176, number=212,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MODF', index=177, number=213,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOOR', index=178, number=214,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CEIL', index=179, number=215,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRUNC', index=180, number=216,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEGREES', index=181, number=217,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RADIANS', index=182, number=218,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIP', index=183, number=219,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISREAL', index=184, number=220,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISCOMPLEX', index=185, number=221,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REAL', index=186, number=222,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IMAG', index=187, number=223,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIX', index=188, number=224,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='I0', index=189, number=225,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SINC', index=190, number=226,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NAN_TO_NUM', index=191, number=227,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISCLOSE', index=192, number=228,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISCLOSE_CONSTANT', index=193, number=229,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIVMOD', index=194, number=231,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ANGLE', index=195, number=232,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_REAL', index=196, number=233,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_REAL_CONSTANT', index=197, number=234,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_IMAG', index=198, number=235,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_IMAG_CONSTANT', index=199, number=236,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TREE_ADD', index=200, number=251,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TREE_MULTIPLY', index=201, number=252,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMSUM', index=202, number=301,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMPROD', index=203, number=302,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROD', index=204, number=303,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUM', index=205, number=304,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAX', index=206, number=305,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MIN', index=207, number=306,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ALL', index=208, number=307,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ANY', index=209, number=308,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEAN_CHUNK', index=210, number=309,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEAN_COMBINE', index=211, number=310,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEAN', index=212, number=311,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMAX', index=213, number=312,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMAX_CHUNK', index=214, number=313,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMAX_COMBINE', index=215, number=314,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMIN', index=216, number=315,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMIN_CHUNK', index=217, number=316,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMIN_COMBINE', index=218, number=317,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANSUM', index=219, number=318,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMAX', index=220, number=319,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMIN', index=221, number=320,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANPROD', index=222, number=321,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMEAN', index=223, number=322,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMEAN_CHUNK', index=224, number=323,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMAX', index=225, number=324,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMAX_CHUNK', index=226, number=325,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMAX_COMBINE', index=227, number=326,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMIN', index=228, number=327,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMIN_CHUNK', index=229, number=328,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMIN_COMBINE', index=230, number=329,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COUNT_NONZERO', index=231, number=330,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOMENT_CHUNK', index=232, number=331,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMOMENT_CHUNK', index=233, number=332,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOMENT_COMBINE', index=234, number=333,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMOMENT_COMBINE', index=235, number=334,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOMENT', index=236, number=335,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMOMENT', index=237, number=336,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VAR', index=238, number=337,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STD', index=239, number=338,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANVAR', index=240, number=339,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANSTD', index=241, number=340,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANCUMSUM', index=242, number=341,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANCUMPROD', index=243, number=342,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESHAPE', index=244, number=401,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SLICE', index=245, number=402,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INDEX', index=246, number=403,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INDEXSETVALUE', index=247, number=404,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONCATENATE', index=248, number=405,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RECHUNK', index=249, number=406,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ASTYPE', index=250, number=407,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRANSPOSE', index=251, number=408,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SWAPAXES', index=252, number=409,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BROADCAST_TO', index=253, number=410,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STACK', index=254, number=411,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WHERE', index=255, number=412,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHOOSE', index=256, number=413,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NONZERO', index=257, number=414,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGWHERE', index=258, number=415,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRAVEL_INDEX', index=259, number=416,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAVEL_MULTI_INDEX', index=260, number=417,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARRAY_SPLIT', index=261, number=418,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQUEEZE', index=262, number=419,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIGITIZE', index=263, number=420,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REPEAT', index=264, number=421,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COPYTO', index=265, number=422,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISIN', index=266, number=423,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSORDOT', index=267, number=501,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DOT', index=268, number=502,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MATMUL', index=269, number=503,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHOLESKY', index=270, number=510,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='QR', index=271, number=511,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVD', index=272, number=512,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LU', index=273, number=513,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOLVE_TRIANGULAR', index=274, number=520,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NORM', index=275, number=530,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFT', index=276, number=601,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFT', index=277, number=602,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFT2', index=278, number=603,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFT2', index=279, number=604,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTN', index=280, number=605,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFTN', index=281, number=606,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFT', index=282, number=607,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRFFT', index=283, number=608,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFT2', index=284, number=609,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRFFT2', index=285, number=610,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFTN', index=286, number=611,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRFFTN', index=287, number=612,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HFFT', index=288, number=613,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IHFFT', index=289, number=614,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTFREQ', index=290, number=615,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTFREQ_CHUNK', index=291, number=616,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFTFREQ', index=292, number=617,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTSHIFT', index=293, number=618,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFTSHIFT', index=294, number=619,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPARSE_MATRIX_DATA_SOURCE', index=295, number=701,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DENSE_TO_SPARSE', index=296, number=702,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPARSE_TO_DENSE', index=297, number=703,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FUSE', index=298, number=801,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENTER', index=299, number=901,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LEAVE', index=300, number=902,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIX_LATEST', index=301, number=903,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IF_ELSE', index=302, number=904,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEXT_ITER', index=303, number=905,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TABLE_COO', index=304, number=1003,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STORE_COO', index=305, number=1004,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FETCH', index=306, number=999999,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=237,
  serialized_end=5070,
)
_sym_db.RegisterEnumDescriptor(_OPERANDDEF_OPERANDTYPE)


_OPERANDDEF_ATTRENTRY = _descriptor.Descriptor(
  name='AttrEntry',
  full_name='OperandDef.AttrEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='OperandDef.AttrEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='OperandDef.AttrEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=183,
  serialized_end=234,
)

_OPERANDDEF = _descriptor.Descriptor(
  name='OperandDef',
  full_name='OperandDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='OperandDef.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='attr', full_name='OperandDef.attr', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='operands', full_name='OperandDef.operands', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='OperandDef.id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_OPERANDDEF_ATTRENTRY, ],
  enum_types=[
    _OPERANDDEF_OPERANDTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=5070,
)

_OPERANDDEF_ATTRENTRY.fields_by_name['value'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_OPERANDDEF_ATTRENTRY.containing_type = _OPERANDDEF
_OPERANDDEF.fields_by_name['attr'].message_type = _OPERANDDEF_ATTRENTRY
_OPERANDDEF.fields_by_name['operands'].message_type = _OPERANDDEF
_OPERANDDEF_OPERANDTYPE.containing_type = _OPERANDDEF
DESCRIPTOR.message_types_by_name['OperandDef'] = _OPERANDDEF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

OperandDef = _reflection.GeneratedProtocolMessageType('OperandDef', (_message.Message,), dict(

  AttrEntry = _reflection.GeneratedProtocolMessageType('AttrEntry', (_message.Message,), dict(
    DESCRIPTOR = _OPERANDDEF_ATTRENTRY,
    __module__ = 'mars.serialize.protos.operand_pb2'
    # @@protoc_insertion_point(class_scope:OperandDef.AttrEntry)
    ))
  ,
  DESCRIPTOR = _OPERANDDEF,
  __module__ = 'mars.serialize.protos.operand_pb2'
  # @@protoc_insertion_point(class_scope:OperandDef)
  ))
_sym_db.RegisterMessage(OperandDef)
_sym_db.RegisterMessage(OperandDef.AttrEntry)


_OPERANDDEF_ATTRENTRY._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorVonmises(operands.Vonmises, TensorRandomOperandMixin):
    __slots__ = '_mu', '_kappa', '_size'
    _input_fields_ = ['_mu', '_kappa']

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorVonmises, self).__init__(_size=size, _state=state, _dtype=dtype,
                                             _gpu=gpu, **kw)

    def __call__(self, mu, kappa, chunk_size=None):
        return self.new_tensor([mu, kappa], None, raw_chunk_size=chunk_size)


def vonmises(random_state, mu, kappa, size=None, chunk_size=None, gpu=None, **kw):
    r"""
    Draw samples from a von Mises distribution.

    Samples are drawn from a von Mises distribution with specified mode
    (mu) and dispersion (kappa), on the interval [-pi, pi].

    The von Mises distribution (also known as the circular normal
    distribution) is a continuous probability distribution on the unit
    circle.  It may be thought of as the circular analogue of the normal
    distribution.

    Parameters
    ----------
    mu : float or array_like of floats
        Mode ("center") of the distribution.
    kappa : float or array_like of floats
        Dispersion of the distribution, has to be >=0.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``mu`` and ``kappa`` are both scalars.
        Otherwise, ``np.broadcast(mu, kappa).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized von Mises distribution.

    See Also
    --------
    scipy.stats.vonmises : probability density function, distribution, or
        cumulative density function, etc.

    Notes
    -----
    The probability density for the von Mises distribution is

    .. math:: p(x) = \frac{e^{\kappa cos(x-\mu)}}{2\pi I_0(\kappa)},

    where :math:`\mu` is the mode and :math:`\kappa` the dispersion,
    and :math:`I_0(\kappa)` is the modified Bessel function of order 0.

    The von Mises is named for Richard Edler von Mises, who was born in
    Austria-Hungary, in what is now the Ukraine.  He fled to the United
    States in 1939 and became a professor at Harvard.  He worked in
    probability theory, aerodynamics, fluid mechanics, and philosophy of
    science.

    References
    ----------
    .. [1] Abramowitz, M. and Stegun, I. A. (Eds.). "Handbook of
           Mathematical Functions with Formulas, Graphs, and Mathematical
           Tables, 9th printing," New York: Dover, 1972.
    .. [2] von Mises, R., "Mathematical Theory of Probability
           and Statistics", New York: Academic Press, 1964.

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> mu, kappa = 0.0, 4.0 # mean and dispersion
    >>> s = mt.random.vonmises(mu, kappa, 1000)

    Display the histogram of the samples, along with
    the probability density function:

    >>> import matplotlib.pyplot as plt
    >>> from scipy.special import i0
    >>> plt.hist(s.execute(), 50, normed=True)
    >>> x = mt.linspace(-mt.pi, mt.pi, num=51)
    >>> y = mt.exp(kappa*mt.cos(x-mu))/(2*mt.pi*i0(kappa))
    >>> plt.plot(x.execute(), y.execute(), linewidth=2, color='r')
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().vonmises(
            handle_array(mu), handle_array(kappa), size=(0,)).dtype

    size = random_state._handle_size(size)
    op = TensorVonmises(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(mu, kappa, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Integral
import contextlib

import numpy as np

from ....operands import IndexSetValue
from ....core import BaseWithKey, Entity
from ...core import TENSOR_TYPE, CHUNK_TYPE
from ..core import TensorOperandMixin
from .core import process_index, get_index_and_shape
from .getitem import TensorIndex


class TensorIndexSetValue(IndexSetValue, TensorOperandMixin):
    def __init__(self, dtype=None, sparse=False, **kw):
        super(TensorIndexSetValue, self).__init__(_dtype=dtype, _sparse=sparse, **kw)

    @contextlib.contextmanager
    def _handle_params(self, inputs, indexes, value):
        """
        TensorIndexSetValue operator is like Index operand, it has additional parameter `indexes` and `value`, all of
        them may be tensor type. As explained in TensorIndex, when indexes and value are not provided, we should get
        from operand itself and replace tensor-liked objects by iterating over inputs.
        """
        if indexes is not None and value is not None:
            indexes_inputs = [ind for ind in indexes if isinstance(ind, TENSOR_TYPE + CHUNK_TYPE)]
            inputs += indexes_inputs
            if isinstance(value, TENSOR_TYPE + CHUNK_TYPE):
                inputs += [value]
        yield inputs

        if indexes is not None:
            self._indexes = indexes
        if value is not None:
            self._value = value
        inputs_iter = iter(self._inputs[1:])
        new_indexes = [next(inputs_iter) if isinstance(index, (BaseWithKey, Entity)) else index
                       for index in self._indexes]
        self._indexes = new_indexes
        if isinstance(self._value, (BaseWithKey, Entity)):
            self._value = next(inputs_iter)

    def new_tensors(self, inputs, shape, **kw):
        indexes = kw.pop('indexes', None)
        value = kw.pop('value', None)
        with self._handle_params(inputs, indexes, value) as mix_inputs:
            return super(TensorIndexSetValue, self).new_tensors(mix_inputs, shape, **kw)

    def new_chunks(self, inputs, shape, **kw):
        indexes = kw.pop('indexes', None)
        value = kw.pop('value', None)
        with self._handle_params(inputs, indexes, value) as mix_inputs:
            return super(TensorIndexSetValue, self).new_chunks(mix_inputs, shape, **kw)

    def __call__(self, a, index, value):
        return self.new_tensor([a], a.shape, indexes=index, value=value)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        value = op.value
        is_value_tensor = isinstance(value, TENSOR_TYPE)

        index_tensor_op = TensorIndex(dtype=tensor.dtype, sparse=op.sparse)
        index_tensor = index_tensor_op.new_tensor([op.input], tensor.shape, indexes=op.indexes).single_tiles()

        nsplits = index_tensor.nsplits
        if any(any(np.isnan(ns) for ns in nsplit) for nsplit in nsplits):
            raise NotImplementedError

        if is_value_tensor:
            value = op.value.rechunk(nsplits).single_tiles()

        chunk_mapping = {c.op.input.index: c for c in index_tensor.chunks}
        out_chunks = []
        for chunk in op.input.chunks:
            index_chunk = chunk_mapping.get(chunk.index)
            if index_chunk is None:
                out_chunks.append(chunk)
                continue

            value_chunk = value.cix[index_chunk.index] if is_value_tensor else value
            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk([chunk], chunk.shape, indexes=index_chunk.op.indexes,
                                           value=value_chunk, index=chunk.index)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors([op.input], tensor.shape, indexes=op.indexes, value=op.value,
                                  chunks=out_chunks, nsplits=op.input.nsplits)


def _setitem(a, item, value):
    from ..base import broadcast_to

    index = process_index(a, item)
    index, shape = get_index_and_shape(a, index)

    for ix in index:
        if not isinstance(ix, (slice, Integral)):
            raise NotImplementedError('Only slice or int supported by now, got {0}'.format(type(ix)))

    if np.isscalar(value):
        value = a.dtype.type(value)
    else:
        value = broadcast_to(value, shape).astype(a.dtype)

    op = TensorIndexSetValue(dtype=a.dtype, sparse=a.issparse())
    ret = op(a, index, value)
    a.data = ret.data



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/value.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/value.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n!mars/serialize/protos/value.proto\"\xa6\x04\n\x05Value\x12\x11\n\x07is_null\x18\x01 \x01(\x08H\x00\x12\x0b\n\x01\x62\x18\x02 \x01(\x08H\x00\x12\x0b\n\x01i\x18\x03 \x01(\x03H\x00\x12\x0b\n\x01\x66\x18\x04 \x01(\x02H\x00\x12\x0b\n\x01s\x18\x05 \x01(\x0cH\x00\x12\x0b\n\x01u\x18\x06 \x01(\tH\x00\x12\x1b\n\x04list\x18\x07 \x01(\x0b\x32\x0b.Value.ListH\x00\x12\x1b\n\x04\x64ict\x18\x08 \x01(\x0b\x32\x0b.Value.DictH\x00\x12\x1d\n\x05slice\x18\t \x01(\x0b\x32\x0c.Value.SliceH\x00\x12\r\n\x03\x61rr\x18\n \x01(\x0cH\x00\x12\x0f\n\x05\x64type\x18\x0b \x01(\x0cH\x00\x12\x19\n\x03key\x18\x0c \x01(\x0b\x32\n.Value.KeyH\x00\x12\x14\n\ndatetime64\x18\r \x01(\x0cH\x00\x12\x15\n\x0btimedelta64\x18\x0e \x01(\x0cH\x00\x1a/\n\x04List\x12\x10\n\x08is_tuple\x18\x01 \x01(\x08\x12\x15\n\x05value\x18\x02 \x03(\x0b\x32\x06.Value\x1an\n\x05Slice\x12\x0f\n\x07is_null\x18\x04 \x01(\x08\x12\x13\n\tstart_val\x18\x01 \x01(\x03H\x00\x12\x12\n\x08stop_val\x18\x02 \x01(\x03H\x01\x12\x12\n\x08step_val\x18\x03 \x01(\x03H\x02\x42\x07\n\x05startB\x06\n\x04stopB\x06\n\x04step\x1a>\n\x04\x44ict\x12\x19\n\x04keys\x18\x01 \x01(\x0b\x32\x0b.Value.List\x12\x1b\n\x06values\x18\x02 \x01(\x0b\x32\x0b.Value.List\x1a\x1e\n\x03Key\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\tB\x07\n\x05valueb\x06proto3')
)




_VALUE_LIST = _descriptor.Descriptor(
  name='List',
  full_name='Value.List',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_tuple', full_name='Value.List.is_tuple', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='Value.List.value', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=324,
  serialized_end=371,
)

_VALUE_SLICE = _descriptor.Descriptor(
  name='Slice',
  full_name='Value.Slice',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_null', full_name='Value.Slice.is_null', index=0,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_val', full_name='Value.Slice.start_val', index=1,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stop_val', full_name='Value.Slice.stop_val', index=2,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step_val', full_name='Value.Slice.step_val', index=3,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='start', full_name='Value.Slice.start',
      index=0, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='stop', full_name='Value.Slice.stop',
      index=1, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='step', full_name='Value.Slice.step',
      index=2, containing_type=None, fields=[]),
  ],
  serialized_start=373,
  serialized_end=483,
)

_VALUE_DICT = _descriptor.Descriptor(
  name='Dict',
  full_name='Value.Dict',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='keys', full_name='Value.Dict.keys', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='values', full_name='Value.Dict.values', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=485,
  serialized_end=547,
)

_VALUE_KEY = _descriptor.Descriptor(
  name='Key',
  full_name='Value.Key',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Value.Key.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='Value.Key.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=549,
  serialized_end=579,
)

_VALUE = _descriptor.Descriptor(
  name='Value',
  full_name='Value',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_null', full_name='Value.is_null', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='b', full_name='Value.b', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='i', full_name='Value.i', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='f', full_name='Value.f', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='s', full_name='Value.s', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='u', full_name='Value.u', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='list', full_name='Value.list', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dict', full_name='Value.dict', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slice', full_name='Value.slice', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='arr', full_name='Value.arr', index=9,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dtype', full_name='Value.dtype', index=10,
      number=11, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key', full_name='Value.key', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='datetime64', full_name='Value.datetime64', index=12,
      number=13, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timedelta64', full_name='Value.timedelta64', index=13,
      number=14, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_VALUE_LIST, _VALUE_SLICE, _VALUE_DICT, _VALUE_KEY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='Value.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=38,
  serialized_end=588,
)

_VALUE_LIST.fields_by_name['value'].message_type = _VALUE
_VALUE_LIST.containing_type = _VALUE
_VALUE_SLICE.containing_type = _VALUE
_VALUE_SLICE.oneofs_by_name['start'].fields.append(
  _VALUE_SLICE.fields_by_name['start_val'])
_VALUE_SLICE.fields_by_name['start_val'].containing_oneof = _VALUE_SLICE.oneofs_by_name['start']
_VALUE_SLICE.oneofs_by_name['stop'].fields.append(
  _VALUE_SLICE.fields_by_name['stop_val'])
_VALUE_SLICE.fields_by_name['stop_val'].containing_oneof = _VALUE_SLICE.oneofs_by_name['stop']
_VALUE_SLICE.oneofs_by_name['step'].fields.append(
  _VALUE_SLICE.fields_by_name['step_val'])
_VALUE_SLICE.fields_by_name['step_val'].containing_oneof = _VALUE_SLICE.oneofs_by_name['step']
_VALUE_DICT.fields_by_name['keys'].message_type = _VALUE_LIST
_VALUE_DICT.fields_by_name['values'].message_type = _VALUE_LIST
_VALUE_DICT.containing_type = _VALUE
_VALUE_KEY.containing_type = _VALUE
_VALUE.fields_by_name['list'].message_type = _VALUE_LIST
_VALUE.fields_by_name['dict'].message_type = _VALUE_DICT
_VALUE.fields_by_name['slice'].message_type = _VALUE_SLICE
_VALUE.fields_by_name['key'].message_type = _VALUE_KEY
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['is_null'])
_VALUE.fields_by_name['is_null'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['b'])
_VALUE.fields_by_name['b'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['i'])
_VALUE.fields_by_name['i'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['f'])
_VALUE.fields_by_name['f'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['s'])
_VALUE.fields_by_name['s'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['u'])
_VALUE.fields_by_name['u'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['list'])
_VALUE.fields_by_name['list'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['dict'])
_VALUE.fields_by_name['dict'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['slice'])
_VALUE.fields_by_name['slice'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['arr'])
_VALUE.fields_by_name['arr'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['dtype'])
_VALUE.fields_by_name['dtype'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['key'])
_VALUE.fields_by_name['key'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['datetime64'])
_VALUE.fields_by_name['datetime64'].containing_oneof = _VALUE.oneofs_by_name['value']
_VALUE.oneofs_by_name['value'].fields.append(
  _VALUE.fields_by_name['timedelta64'])
_VALUE.fields_by_name['timedelta64'].containing_oneof = _VALUE.oneofs_by_name['value']
DESCRIPTOR.message_types_by_name['Value'] = _VALUE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), dict(

  List = _reflection.GeneratedProtocolMessageType('List', (_message.Message,), dict(
    DESCRIPTOR = _VALUE_LIST,
    __module__ = 'mars.serialize.protos.value_pb2'
    # @@protoc_insertion_point(class_scope:Value.List)
    ))
  ,

  Slice = _reflection.GeneratedProtocolMessageType('Slice', (_message.Message,), dict(
    DESCRIPTOR = _VALUE_SLICE,
    __module__ = 'mars.serialize.protos.value_pb2'
    # @@protoc_insertion_point(class_scope:Value.Slice)
    ))
  ,

  Dict = _reflection.GeneratedProtocolMessageType('Dict', (_message.Message,), dict(
    DESCRIPTOR = _VALUE_DICT,
    __module__ = 'mars.serialize.protos.value_pb2'
    # @@protoc_insertion_point(class_scope:Value.Dict)
    ))
  ,

  Key = _reflection.GeneratedProtocolMessageType('Key', (_message.Message,), dict(
    DESCRIPTOR = _VALUE_KEY,
    __module__ = 'mars.serialize.protos.value_pb2'
    # @@protoc_insertion_point(class_scope:Value.Key)
    ))
  ,
  DESCRIPTOR = _VALUE,
  __module__ = 'mars.serialize.protos.value_pb2'
  # @@protoc_insertion_point(class_scope:Value)
  ))
_sym_db.RegisterMessage(Value)
_sym_db.RegisterMessage(Value.List)
_sym_db.RegisterMessage(Value.Slice)
_sym_db.RegisterMessage(Value.Dict)
_sym_db.RegisterMessage(Value.Key)


# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .nonzero import nonzero


def flatnonzero(a):
    """
    Return indices that are non-zero in the flattened version of a.

    This is equivalent to a.ravel().nonzero()[0].

    Parameters
    ----------
    a : Tensor
        Input tensor.

    Returns
    -------
    res : Tensor
        Output tensor, containing the indices of the elements of `a.ravel()`
        that are non-zero.

    See Also
    --------
    nonzero : Return the indices of the non-zero elements of the input tensor.
    ravel : Return a 1-D tensor containing the elements of the input tensor.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(-2, 3)
    >>> x.execute()
    array([-2, -1,  0,  1,  2])
    >>> mt.flatnonzero(x).execute()
    array([0, 1, 3, 4])

    Use the indices of the non-zero elements as an index array to extract
    these elements:

    >>> x.ravel()[mt.flatnonzero(x)].execute()  # TODO(jisheng): accomplish this after fancy indexing is supported

    """
    from ..base import ravel
    return nonzero(ravel(a))[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..datasource import tensor as astensor
from .solve import solve


def inv(a):
    """
    Compute the (multiplicative) inverse of a matrix.
    Given a square matrix `a`, return the matrix `ainv` satisfying
    ``dot(a, ainv) = dot(ainv, a) = eye(a.shape[0])``.
    Parameters
    ----------
    a : (..., M, M) array_like
        Matrix to be inverted.
    Returns
    -------
    ainv : (..., M, M) ndarray or matrix
        (Multiplicative) inverse of the matrix `a`.
    Raises
    ------
    LinAlgError
        If `a` is not square or inversion fails.
    Notes
    -----
    .. versionadded:: 1.8.0
    Broadcasting rules apply, see the `numpy.linalg` documentation for
    details.
    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = np.array([[1., 2.], [3., 4.]])
    >>> ainv = mt.linalg.inv(a)
    >>> mt.allclose(mt.dot(a, ainv), mt.eye(2)).execute()
    True
    >>> mt.allclose(mt.dot(ainv, a), mt.eye(2)).execute()
    True
    >>> ainv.execute()
    array([[ -2. ,  1. ],
           [ 1.5, -0.5]])
    """
    from ..datasource import eye

    a = astensor(a)
    return solve(a, eye(a.shape[0], chunk_size=a.params.raw_chunk_size))
